import { camelCase } from "change-case";
import dedent from "dedent";
import fs from "fs-extra";
import path from "node:path";
import { format } from "prettier";
import { resources } from "sidebar/shared.js";

const referencePackages = [
  "@account-kit/core",
  "@account-kit/infra",
  "@account-kit/react",
  "@account-kit/signer",
  "@account-kit/smart-contracts",
  "@account-kit/react-native",
  "@aa-sdk/core",
  "@aa-sdk/ethers",
];

const generateItems = (
  items: string[],
  referencePath: string,
  location: "functions" | "hooks" | "components" | `classes/${string}`
) => {
  return items.map((file) => {
    const functionName = file.split(".")[0];
    return {
      text: functionName,
      link: `${referencePath.replace(".", "")}/${location}/${functionName}`,
    };
  });
};

// Define the hook groupings
const hookGroupings = {
  Authentication: [
    "useAuthModal",
    "useAddPasskey",
    "useAuthenticate",
    "useConnect",
    "useExportAccount",
    "useLogout",
    "useSigner",
    "useSignerStatus",
    "useUser",
  ],
  "Account Client": ["useAccount", "useSmartAccountClient"],
  "Smart Account Actions": [
    "useSendUserOperation",
    "useSignMessage",
    "useSignTypedData",
    "useWaitForUserOperationTransaction",
    "useDropAndReplaceUserOperation",
    "useClientActions",
    "useChain",
  ],
  "Bundler/RPC Client": ["useBundlerClient"],
  Utilities: [
    "useAlchemyAccountContext",
    "useUiConfig",
    "useConnection",
    "useAuthError",
  ],
};

referencePackages.forEach(async (pkg) => {
  const cleanPkg = pkg.replace("@", "");
  const referencePath = `./reference/${cleanPkg}`;
  const functionFiles = await fs
    .readdir(path.resolve("./pages", referencePath, "./functions"))
    .catch(() => []);
  const hookFiles = await fs
    .readdir(path.resolve("./pages", referencePath, "./hooks"))
    .catch(() => []);
  const componentFiles = await fs
    .readdir(path.resolve("./pages", referencePath, "./components"))
    .catch(() => []);
  const classes = await fs
    .readdir(path.resolve("./pages", referencePath, "./classes"))
    .catch(() => []);
  const classFiles = (
    await Promise.all(
      classes.map(async (c) => {
        const classLocation = `./classes/${c}`;
        const functions = await fs
          .readdir(path.resolve("./pages", referencePath, classLocation))
          .catch(() => [] as string[]);
        if (!functions.length) return;

        return {
          className: c,
          functions,
        };
      })
    )
  ).filter((x) => x != null);

  const functionItems = generateItems(
    functionFiles,
    referencePath,
    "functions"
  );
  const hookItems = generateItems(hookFiles, referencePath, "hooks");

  // Group hooks according to the specified categories
  const groupedHookItems = Object.entries(hookGroupings)
    .map(([group, hooks]) => {
      const groupItems = hookItems.filter((hook) => hooks.includes(hook.text));
      if (groupItems.length) {
        return {
          text: group,
          items: groupItems,
        };
      }
      return null;
    })
    .filter((x) => x != null);

  const componentItems = generateItems(
    componentFiles,
    referencePath,
    "components"
  );
  const classItems = classFiles.map((c) => {
    const classFunctions = generateItems(
      c.functions,
      referencePath,
      `classes/${c.className}`
    );

    return {
      text: c.className,
      items: classFunctions,
    };
  });

  const componentsSection = componentItems.length
    ? { text: "Components", items: componentItems }
    : undefined;
  const functionsSection = functionItems.length
    ? { text: "Functions", items: functionItems }
    : undefined;
  const classesSection = classItems.length
    ? { text: "Classes", items: classItems }
    : undefined;

  const sidebarConst = `${camelCase(
    cleanPkg.split("/").join("-")
  )}ReferenceSidebar`;
  const sidebarContent = dedent`
  // This file is autogenerated, see scripts/prebuild.ts
  import { SidebarItem } from "vocs";

  export const ${sidebarConst}: SidebarItem[] = [
    {
        text: "${pkg}",
        items: [{
            text: "Overview",
            link: "${referencePath.replace(".", "")}"
        }],
    },
    ${[componentsSection, ...groupedHookItems, functionsSection, classesSection]
      .filter((x) => x != null)
      .map((x) => JSON.stringify(x, null, 2))
      .join(",\n")},
    ${JSON.stringify(resources, null, 2)}
  ];
  `;

  fs.outputFileSync(
    path.resolve("./sidebar", `${referencePath}.ts`),
    format(sidebarContent, { parser: "typescript" })
  );
});
